@@ -34,11 +34,7 @@
 #include "IIOP.h"
 #include "IIOP-private.h"
 
-#ifdef HAVE_LIMITED_WRITEV
-#define writev g_writev
-#endif
-
-/* type defs */
+#define MAX_LIMITED_IOVECS 16
 
 #ifdef        __GNUC__
 #define       PACKED __attribute__((packed))
@@ -172,76 +168,92 @@ giop_send_buffer_new(void)
 gint
 giop_send_buffer_write(GIOPSendBuffer *send_buffer)
 {
-  gulong nvecs;
-  glong res, sum, t;
-  struct iovec *curvec;
-  int fd;
-  GIOPConnection *cnx;
-  gint retval = -1;
+	gulong nvecs;
+	glong res, length_to_write;
+	struct iovec *curvec;
+	int fd;
+	gboolean set_blocking;
+	GIOPConnection *cnx;
+	gint retval = -1;
 
-  cnx = GIOP_MESSAGE_BUFFER(send_buffer)->connection;
-  if(!cnx->is_valid)
-    return -1;
+	cnx = GIOP_MESSAGE_BUFFER (send_buffer)->connection;
+	if (!cnx->is_valid)
+		return -1;
 
-  fd = GIOP_CONNECTION_GET_FD(cnx);
-  nvecs = GIOP_MESSAGE_BUFFER(send_buffer)->iovecs->len;
-  curvec = (struct iovec *)GIOP_MESSAGE_BUFFER(send_buffer)->iovecs->data;
+	fd = GIOP_CONNECTION_GET_FD (cnx);
+	nvecs = GIOP_MESSAGE_BUFFER (send_buffer)->iovecs->len;
+	curvec = (struct iovec *) GIOP_MESSAGE_BUFFER (send_buffer)->iovecs->data;
 
-#if defined(ORBIT_DEBUG) && 0
-  g_print("Message of length %d looks like:\n",
-	  GIOP_MESSAGE_BUFFER(send_buffer)->message_header.message_size);
-  for(sum = 0, t = 2; t < nvecs; t++) {
+#if defined (ORBIT_DEBUG) && 0
+	{
+		gulong sum, t;
 
-    sum += curvec[t].iov_len;
-    g_print("    [%p, %d]: %d\n", curvec[t].iov_base, curvec[t].iov_len,
-	    sum);
-  }
+		g_print ("Message of length %d looks like:\n",
+			 GIOP_MESSAGE_BUFFER (send_buffer)->message_header.message_size);
+		for (sum = 0, t = 0; t < nvecs; t++) {
+			sum += curvec[t].iov_len;
+			g_print ("    [%p, %d]: cumul. %ld   - vec idx %d\n",
+				 curvec[t].iov_base,
+				 curvec[t].iov_len, sum, t);
+		}
+	}
 #endif
-  res = writev(fd, curvec, nvecs);
+	length_to_write = sizeof (GIOPMessageHeader)+ 
+		GIOP_MESSAGE_BUFFER(send_buffer)->message_header.message_size;
+	set_blocking = FALSE;
+	do {
+  
+		do {
+#ifdef HAVE_LIMITED_WRITEV
+/*			fprintf (stderr, "ltw %ld write %d vecs, first of length %d ",
+			length_to_write, nvecs, (int) curvec->iov_len);*/
+			res = writev (fd, curvec, MIN (nvecs, MAX_LIMITED_IOVECS));
+/*			fprintf (stderr, "wrote %ld bytes [%d]\n", res, (int) errno); */
+#else
+			res = writev (fd, curvec, nvecs);
+#endif
+		} while (res < 0 && errno == EINTR);
 
-  sum = (GIOP_MESSAGE_BUFFER(send_buffer)->message_header.message_size + sizeof(GIOPMessageHeader));
-  if(res < sum) {
-    if(res < 0) {
-      if(errno != EAGAIN) {
-	giop_main_handle_connection_exception(cnx);
-	goto out;
-      }
+		if (res < 0) {
+			if (errno != EAGAIN) {
+				giop_main_handle_connection_exception (cnx);
+				goto out;
+			} else
+				res = 0;
+		}
 
-      res = 0;
-    }
+		length_to_write -= res;
 
-    /* wrote 7, iovecs 3, 2, 2, 4:
-       0 + 3 !> 7
-       3 + 2 !> 7
-       5 + 2 !> 7
-     */
+		/* Move along the vector array */
+		if (length_to_write > 0) {
+			gulong t;
 
-    for(t = 0; ; t += curvec->iov_len, curvec++, nvecs--) {
-      if((t + curvec->iov_len) > res)
-	break;
-    }
-    if((res - t) > 0) {
-      curvec->iov_len -= (res - t);
-      curvec->iov_base = (gpointer)((char *)curvec->iov_base + (res - t));
-    }
+			if (!set_blocking) {
+				set_blocking = TRUE;
+				fcntl (fd, F_SETFL, fcntl (fd, F_GETFL, 0) & ~O_NONBLOCK);
+			}
 
-    fcntl(fd, F_SETFL, fcntl(fd, F_GETFL, 0) & ~O_NONBLOCK);
+			for (t = 0; (t + curvec->iov_len) <= res; ) {
+				t += curvec->iov_len;
+				curvec++, nvecs--;
+			}
+			/* t is now the count of the bytes in completely written iovecs */
 
-    t = writev(fd, curvec, nvecs);
+			if ((res - t) > 0) {
+				curvec->iov_len -= (res - t);
+				curvec->iov_base = ((guchar *)curvec->iov_base + (res - t));
+			}
+		}
 
-    fcntl(fd, F_SETFL, fcntl(fd, F_GETFL, 0) | O_NONBLOCK);
+	} while (length_to_write > 0);
 
-    if((t < 0) || ((res + t) < sum)) {
-	giop_main_handle_connection_exception(cnx);
-	goto out;
-    }
-  }
+	if (set_blocking)
+		fcntl (fd, F_SETFL, fcntl (fd, F_GETFL, 0) | O_NONBLOCK);
 
-  retval = 0;
+	retval = 0;
 
  out:
-
-  return retval;
+	return retval;
 }
 
 static GIOPSendBuffer *
@@ -809,6 +821,65 @@ giop_recv_buffer_use(GIOPConnection *connection)
   return retval;
 }
 
+#ifdef GIOP_INTERNAL_DEBUG
+static void
+dump (FILE *out, guint8 const *ptr, guint32 len, guint32 offset)
+{
+	guint32 lp,lp2;
+	guint32 off;
+
+	for (lp = 0;lp<(len+15)/16;lp++) {
+		fprintf (out, "0x%4x ", offset + lp * 16);
+		for (lp2=0;lp2<16;lp2++) {
+			off = lp2 + (lp<<4);
+			off<len?fprintf (out, "%2x ", ptr[off]):fprintf (out, "XX ");
+		}
+		fprintf (out, "| ");
+		for (lp2=0;lp2<16;lp2++) {
+			off = lp2 + (lp<<4);
+			fprintf (out, "%c", off<len?(ptr[off]>'!'&&ptr[off]<127?ptr[off]:'.'):'*');
+		}
+		if (lp == 0)
+			fprintf (out, " --- \n");
+		else
+			fprintf (out, "\n");
+	}
+}
+
+/*
+void
+giop_dump_send (GIOPSendBuffer *send_buffer)
+{
+	gulong nvecs;
+	struct iovec *curvec;
+	guint32 offset = 0;
+
+	g_return_if_fail (send_buffer != NULL);
+
+	nvecs = send_buffer->num_used;
+	curvec = (struct iovec *) send_buffer->iovecs;
+
+	fprintf (stderr, "Outgoing IIOP data:\n");
+	while (nvecs-- > 0) {
+		dump (stderr, curvec->iov_base, curvec->iov_len, offset);
+		offset += curvec->iov_len;
+		curvec++;
+	}
+} */
+
+void
+giop_dump_recv (GIOPRecvBuffer *recv_buffer)
+{
+	g_return_if_fail (recv_buffer != NULL);
+
+	fprintf (stderr, "Incoming IIOP data\n");
+
+	dump (stderr, recv_buffer->message_body,
+	      GIOP_MESSAGE_BUFFER (recv_buffer)->message_header.message_size +
+	      sizeof (GIOPMessageHeader), 0);
+}
+#endif /* GIOP_INTERNAL_DEBUG */
+
 GIOPRecvBuffer *
 giop_recv_message_buffer_use(GIOPConnection *connection)
 {
@@ -829,6 +900,10 @@ giop_recv_message_buffer_use(GIOPConnection *connection)
 
   if(!retval) return NULL;
 
+#ifdef GIOP_INTERNAL_DEBUG
+  g_warning ("Recv message buffer use");
+#endif
+
   do {
     switch(retval->state) {
     case GIOP_MSG_READING_HEADER:
@@ -865,6 +940,7 @@ giop_recv_message_buffer_use(GIOPConnection *connection)
 	if(strncmp(GIOP_MESSAGE_BUFFER(retval)->message_header.magic, "GIOP", 4)
 	   || GIOP_MESSAGE_BUFFER(retval)->message_header.GIOP_version[0] != 1)
 	  goto errout;
+
 	if(GIOP_MESSAGE_BUFFER(retval)->message_header.message_size == 0
 	   && GIOP_MESSAGE_BUFFER(retval)->message_header.message_type != GIOP_CLOSECONNECTION) {
 	  g_warning("Unexpected 0-length IIOP message");
@@ -891,7 +967,7 @@ giop_recv_message_buffer_use(GIOPConnection *connection)
 	retval->message_body = g_malloc(message_size+sizeof(GIOPMessageHeader)+4);
 	/* XXX1 This is a lame hack to work with the fact that
 	   alignment is relative to the MessageHeader, not the RequestHeader */
-	retval->cur = retval->message_body + 12;
+	retval->cur = (gpointer)(((char *)retval->message_body) + 12);
 	retval->state = GIOP_MSG_READING_BODY;
 	retval->left_to_read = message_size;
 	break;
@@ -912,6 +988,11 @@ giop_recv_message_buffer_use(GIOPConnection *connection)
       retval = NULL;
     }
   } while(retval && retval->state != GIOP_MSG_READY);
+
+#ifdef GIOP_INTERNAL_DEBUG
+  g_warning ("Recv message buffer use");
+  giop_dump_recv (retval);
+#endif /* GIOP_INTERNAL_DEBUG */
   
   return retval;
 
@@ -1178,13 +1259,20 @@ giop_recv_reply_decode_message(GIOPRecvBuffer *buf)
   */
   int i;
 
+#ifdef GIOP_INTERNAL_DEBUG
+  g_warning ("recv_reply_decode_message");
+#endif
+
   buf->message.u.reply.service_context._maximum = 0;
   if(giop_msg_conversion_needed(GIOP_MESSAGE_BUFFER(buf))) 
     {
       GET_ULONG(buf->message.u.reply.service_context._length);
 /* XXX bad hardcoded hack until someone gives a "right answer" to how to
 solve this problem */
-      if(buf->message.u.reply.service_context._length > 128) return -1;
+      if(buf->message.u.reply.service_context._length > 128) {
+	      g_warning ("length '0x%x' > 128", buf->message.u.reply.service_context._length);
+	      return -1;
+      }
       buf->message.u.reply.service_context._buffer =
 	g_new(IOP_ServiceContext, buf->message.u.reply.service_context._length);
 
@@ -1207,7 +1295,10 @@ solve this problem */
       GET_ULONG_NC(&buf->message.u.reply.service_context._length);
 /* XXX bad hardcoded hack until someone gives a "right answer" to how to
 solve this problem */
-      if(buf->message.u.reply.service_context._length > 128) return -1;
+      if(buf->message.u.reply.service_context._length > 128) {
+	      g_warning ("length '0x%x' > 128", buf->message.u.reply.service_context._length);
+	      return -1;
+      }
       buf->message.u.reply.service_context._buffer =
 	g_new(IOP_ServiceContext, buf->message.u.reply.service_context._length);
 
@@ -1225,7 +1316,7 @@ solve this problem */
       GET_ULONG_NC(&buf->message.u.reply.reply_status);
     }
 
-#if 0
+#ifdef GIOP_INTERNAL_DEBUG
   g_message("[%d] Received reply %d size %d to request %d",
 	    getpid(),
 	    buf->message.u.reply.reply_status,
@@ -1356,7 +1447,7 @@ giop_recv_request_decode_message(GIOPRecvBuffer *buf)
       buf->cur = ((guchar *)buf->cur) + buf->message.u.request.requesting_principal._length;
     }
 
-#if 0
+#ifdef GIOP_INTERNAL_DEBUG
   g_message("[%d] Received request %s size %d ID %d",
 	    getpid(),
 	    buf->message.u.request.operation,
